; Atari executable booter by Fatih Aygun

.include "atari.inc"

.import	__BUFFER_START__
.import __BOOT_START__

buffer = __BUFFER_START__

BOOT_ERROR_EOF = 'e'
BOOT_ERROR_FORMAT = 'f'
BOOT_ERROR_DISK = 'd'
BOOT_ERROR_CHUNK = 'c'

; -------------------------------------------------------------------------

; Disk boot header

.segment "HEADER"

	.byte	0
	.byte	3
	.word	__BOOT_START__
	.word	init

rtsadr:
	rts		; Return from the immediate entry

; -------------------------------------------------------------------------

; Executable size (24-bit big endian)

.segment "PARAMS"

file_size:
 	.byte	0
 	.byte	0
	.byte	0

; -------------------------------------------------------------------------

; Global data

.data
	first_chunk:	.byte	1		; Is this the first chunk?
	buffer_offset:	.byte	128		; Buffer offset
	sector:		.word	4		; Next sector to be read

.bss
	last_word:	.res	2		; Last word read
	chunk_address:	.res	2		; Chunk address

.zeropage
	count:		.res	2		; Number of bytes to read
	load_address:	.res	2		; Address to read into

.code

; -------------------------------------------------------------------------

; Read a word in 'last_word'

.proc read_word
	lda	#2
	sta	count
	lda	#0
	sta	count + 1
	lda	#<last_word
	sta	load_address
	lda	#>last_word
	sta	load_address + 1
	; Fall through
.endproc

; -------------------------------------------------------------------------

; Read 'count' bytes in 'load_address'

.proc read_chunk
	; Decrement 'file_size'
	lda	file_size
	sec
	sbc	count
	sta	file_size
	lda	file_size + 1
	sbc	count + 1
	sta	file_size + 1
	lda	file_size + 2
	sbc	#0
	sta	file_size + 2
	bcs	noerr

	; Not enough bytes to read
	pla
	pla
	jmp	(RUNAD)

noerr:
	; If buffer is not empty, use those bytes
	ldy	#0
	ldx	buffer_offset
	bpl	loop

	; Buffer is empty, fill it
fill:	jsr	fill_buffer

	; Copy the buffer
loop:	lda	count
	bne	next
	lda	count + 1
	beq	done
	dec	count + 1
next:	dec	count

	lda	buffer,x
	sta	(load_address),y
	iny
	bne	incx
	inc	load_address + 1
incx:	inx
	bpl	loop
	bmi	fill

done:	stx	buffer_offset
	rts
.endproc

; -------------------------------------------------------------------------

; Error handler

.proc error
	ldy	#0
	sta	(SAVMSC),y
	jmp	error
.endproc

; -------------------------------------------------------------------------

; Actual loader

.proc init
	; Reset RUNAD vector
	lda	#<rtsadr
	sta	RUNAD
	lda	#>rtsadr
	sta	RUNAD + 1

loop:	; Skip the $FFFF header
	jsr	read_word
	lda	last_word
	and	last_word + 1
	cmp	#$FF
	beq	hdrok

	; No header, if this is the first chunk, this is an error
	lsr	first_chunk
	bcc	store
	lda	#BOOT_ERROR_FORMAT
	jmp	error

	; Read chunk load address
hdrok:	lsr	first_chunk
	jsr	read_word

	; Store chunk load address
store:	lda	last_word
	sta	chunk_address
	lda	last_word + 1
	sta	chunk_address + 1

	; Read chunk size
	jsr	read_word
	inc	last_word
	bne	@skip
	inc	last_word + 1
@skip:	lda	last_word
	sec
	sbc	chunk_address
	sta	count
	lda	last_word + 1
	sbc	chunk_address + 1
	sta	count + 1
	bcs	@noerr
	lda	#BOOT_ERROR_CHUNK
	jmp	error
@noerr:
	; Start from chunk address
	lda	chunk_address
	sta	load_address
	lda	chunk_address + 1
	sta	load_address + 1

	; Reset INITAD vector
	lda	#<rtsadr
	sta	INITAD
	lda	#>rtsadr
	sta	INITAD + 1

	; Load the chunk into memory
	jsr	read_chunk

	; Call INITAD vector
	jsr	goinit

while:	; Loop while there are bytes to read
	lda	file_size
	ora	file_size + 1
	ora	file_size + 2
	bne	loop

	; Done, jump to RUNAD vector
	jmp	(RUNAD)

	; Jump to INITAD vector
goinit:	jmp	(INITAD)
.endproc

; -------------------------------------------------------------------------

; Fill the buffer with a new sector

.proc fill_buffer
	tya
	pha
	lda	#<buffer
	sta	DBUFLO
	lda	#>buffer
	sta	DBUFHI
	lda	#$31
	sta	DDEVIC
	lda	#1
	sta	DUNIT
	lda	#$40
	sta	DSTATS
	lda	#SIO_READ
	sta	DCOMND
	lda	sector
	sta	DAUX1
	lda	sector + 1
	sta	DAUX2
	lda	#128
	sta	DBYTLO
	lda	#0
	sta	DBYTHI
	jsr	SIOV
	bpl	noerr
	lda	#BOOT_ERROR_DISK
	jmp	error
noerr:	inc	sector
	bne	skip
	inc	sector + 1
skip:	ldx	#0
	pla
	tay
	rts
.endproc
