#include "autoboot.h"
#include "aspeqtsettings.h"

AutoBoot::~AutoBoot()
{
    close();
}

void AutoBoot::handleCommand(quint8 command, quint16 aux)
{
    switch (command) {
        case 0x3F:
            {    /* Get high speed index */
                if (!sio->port()->writeCommandAck()) {
                    return;
                }
                sio->port()->writeComplete();
                QByteArray speed(1, 0);
                switch (aspeqtSettings->maximumSerialPortSpeed()) {
                    case 0:
                        speed[0] = 0x28;    // 19200
                        break;
                    case 1:
                        speed[0] = 0x10;    // 38400
                        break;
                    case 2:
                    default:
                        speed[0] = 0x08;    // 57600
                        break;
                }
                sio->port()->writeDataFrame(speed);
//                emitUiMessage(QString("[%1] Speed poll.").arg(deviceName()), MessageType::NormalMessage);
                break;
            }
        case 0x52:
            {    /* Read sector */
                if (aux >= 1 && aux <= sectorCount) {
                    if (!sio->port()->writeCommandAck()) {
                        return;
                    }
                    QByteArray data;
                    if (readSector(aux, data)) {
                        sio->port()->writeComplete();
                        sio->port()->writeDataFrame(data);
//                        emitUiMessage(QString("[%1] Read sector %2 (%3 bytes).")
//                                       .arg(deviceName())
//                                       .arg(aux)
//                                       .arg(data.size()), MessageType::NormalMessage);
                    } else {
                        sio->port()->writeError();
                        emitUiMessage(QString("[%1] Read sector %2 failed.")
                                       .arg(deviceName())
                                       .arg(aux), MessageType::ErrorMessage);
                    }
                } else {
                    sio->port()->writeCommandNak();
                    emitUiMessage(QString("[%1] Read sector %2 NAKed.")
                                   .arg(deviceName())
                                   .arg(aux), MessageType::WarningMessage);
                }
                break;
            }
        case 0x53:
            {    /* Get status */
                if (!sio->port()->writeCommandAck()) {
                    return;
                }

                QByteArray status(4, 0);
                status[0] = 8;
                status[3] = 1;
                sio->port()->writeComplete();
                sio->port()->writeDataFrame(status);
//                emitUiMessage(QString("[%1] Get status.")
//                               .arg(deviceName()), MessageType::NormalMessage);
                break;
            }
        default:        /* Unknown command */
            sio->port()->writeCommandNak();
            emitUiMessage(QString("[%1] command: $%2, aux: $%3 NAKed.")
                           .arg(deviceName())
                           .arg(command, 2, 16, QChar('0'))
                           .arg(aux, 4, 16, QChar('0')), MessageType::WarningMessage);
            break;
    }
}

bool AutoBoot::open(const QString &fileName)
{
    close();
    QFile boot(":/binaries/atari/autoboot/autoboot.bin");
    if (!boot.open(QFile::ReadOnly)) {
        emitUiMessage(tr("Cannot open resource ':/binaries/atari/autoboot/autoboot.bin': %1")
                       .arg(boot.errorString()), MessageType::ErrorMessage);
        return false;
    }
    bootSectors[0] = boot.read(128);
    bootSectors[1] = boot.read(128);
    bootSectors[2] = boot.read(128);
    if (bootSectors[0].count() + bootSectors[0].count() + bootSectors[0].count() != 384) {
        emitUiMessage(tr("Cannot read from resource ':/binaries/atari/autoboot/autoboot.bin': %1")
                       .arg(boot.errorString()), MessageType::ErrorMessage);
        return false;
    }
    boot.close();
    file.setFileName(fileName);
    if (!file.open(QFile::ReadOnly)) {
        emitUiMessage(tr("Cannot open file '%1': %2")
                       .arg(fileName)
                       .arg(boot.errorString()), MessageType::ErrorMessage);
        return false;
    }
    qint64 size = file.size();
    if (size > 65532 * 128) {
        emitUiMessage(tr("Cannot open file '%1': File too large.")
                       .arg(fileName), MessageType::ErrorMessage);
        return false;
    }
    bootSectors[0][7] = size % 256;
    bootSectors[0][8] = (size / 256) % 256;
    bootSectors[0][9] = size / 65536;
    sectorCount = 3 + (file.size() + 127) / 128;
    return true;
}

void AutoBoot::close()
{
    if (file.isOpen()) {
        file.close();
    }
}

bool AutoBoot::readSector(quint16 sector, QByteArray &data)
{
    if (sector <= 3) {
        data = bootSectors[sector - 1];
        emit sectorRead(sector, sectorCount);
        return true;
    }
    sector -= 4;
    if (!file.seek(sector * 128)) {
        emitUiMessage(tr("Cannot seek in file '%1': %2.")
                       .arg(file.fileName())
                       .arg(file.errorString()), MessageType::ErrorMessage);
        return false;
    }
    qint64 bytesToRead = file.bytesAvailable();
    if (bytesToRead > 128 ) {
        bytesToRead = 128;
    }
    data = file.read(bytesToRead);
    if (data.size() != (int)bytesToRead) {
        emitUiMessage(tr("Cannot read from '%1': %2")
                       .arg(file.fileName())
                       .arg(file.errorString()), MessageType::ErrorMessage);
        return false;
    }
    data.resize(128);
    emit sectorRead(sector + 4, sectorCount);
    return true;
}

QString AutoBoot::deviceName()
{
    return "AutoBoot";
}
