#include "serialport.h"
#include "sioworker.h"
#include "atarisio.h"

#include <QTime>

#include <string.h>
#include <fcntl.h>
#include <stropts.h>
#include <termio.h>
#include <errno.h>
#include <sys/types.h>

AbstractSerialPortBackend::AbstractSerialPortBackend(QObject *parent)
    : QObject(parent)
{
}

AbstractSerialPortBackend::~AbstractSerialPortBackend()
{
}

StandardSerialPortBackend::StandardSerialPortBackend(QObject *parent)
    : AbstractSerialPortBackend(parent)
{
    mHandle = -1;
}

StandardSerialPortBackend::~StandardSerialPortBackend()
{
    if (isOpen()) {
        close();
    }
}

QString StandardSerialPortBackend::defaultPortName()
{
    return QString("/dev/ttyS0");
}

bool StandardSerialPortBackend::open(QString name, SerialLine method)
{
    if (isOpen()) {
        close();
    }

    mHandle = ::open(name.toLocal8Bit().constData(), O_RDWR | O_NOCTTY | O_NDELAY);

    if (mHandle < 0) {
        emitUiMessage(tr("Cannot open serial port '%1': %2")
                       .arg(name, lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }

    int status;
    if (!ioctl(mHandle, TIOCMGET, &status) < 0) {
        emitUiMessage(tr("Cannot clear DTR and RTS lines in serial port '%1': %2").arg(name, lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    status = status & ~(TIOCM_DTR & TIOCM_RTS);
    if (!ioctl(mHandle, TIOCMSET, status)) {
        emitUiMessage(tr("Cannot clear DTR and RTS lines in serial port '%1': %2").arg(name, lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }

    mMethod = method;
    mCanceled = false;

    return true;
}

bool StandardSerialPortBackend::isOpen()
{
    return mHandle >= 0;
}

void StandardSerialPortBackend::close()
{
    cancel();
    if (::close(mHandle)) {
        emitUiMessage(tr("Cannot close serial port: %1")
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
    }
    mHandle = -1;
}

void StandardSerialPortBackend::cancel()
{
    mCanceled = true;
}

void StandardSerialPortBackend::setSpeed(uint speed)
{
    termios tios;

    tcgetattr(mHandle, &tios);
    tios.c_cflag &= ~CSTOPB;
    cfmakeraw(&tios);
    switch (speed) {
        case 600:
            cfsetispeed(&tios, B600);
            cfsetospeed(&tios, B600);
            break;
        case 19200:
            cfsetispeed(&tios, B19200);
            cfsetospeed(&tios, B19200);
            break;
        case 38400:
            cfsetispeed(&tios, B38400);
            cfsetospeed(&tios, B38400);
            break;
        case 57600:
            cfsetispeed(&tios, B57600);
            cfsetospeed(&tios, B57600);
            break;
        default:
            emitUiMessage(tr("Cannot set serial port speed to %1: %2")
                           .arg(speed)
                           .arg(lastErrorMessage()), MessageType::ErrorMessage);
            return;
    }

    /* Set serial port state */
    if (tcsetattr(mHandle, TCSANOW, &tios) != 0) {
        emitUiMessage(tr("Cannot set serial port speed to %1: %2")
                       .arg(speed)
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
        return;
    }

    emit speedChanged(speed);
    emitUiMessage(QString("Serial port speed set to %1.").arg(speed), MessageType::ImportantMessage);
    mSpeed = speed;
}

uint StandardSerialPortBackend::speed()
{
    return mSpeed;
}

QByteArray StandardSerialPortBackend::readCommandFrame()
{
    QByteArray data;
    int mask;

    switch (mMethod) {
    case Ri:
        mask = TIOCM_RI;
        break;
    case Dsr:
        mask = TIOCM_DSR;
        break;
    default:
    case Cts:
        mask = TIOCM_CTS;
        break;
    }

    int status;

    do {
        if (ioctl(mHandle, TIOCMGET, &status) < 0) {
            emitUiMessage(tr("Cannot retrieve serial port status: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
            return data;
        }
        if (!(status & mask)) {
            QThread::yieldCurrentThread();
        }
    } while (!(status & mask) && !mCanceled);

    if (mCanceled) {
        return data;
    }

    if (tcflush(mHandle, TCIFLUSH) != 0) {
        emitUiMessage(tr("Cannot clear serial port read buffer: %1")
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
        return data;
    }

    data = readDataFrame(4, false);

    if (!data.isEmpty()) {
        do {
            if (ioctl(mHandle, TIOCMGET, &status) < 0) {
                emitUiMessage(tr("Cannot retrieve serial port status: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
                return data;
            }
        } while ((status & mask) && !mCanceled);
    }

    return data;
}

QByteArray StandardSerialPortBackend::readDataFrame(uint size, bool verbose)
{
    QByteArray data = readRawFrame(size + 1, verbose);
    if (data.isEmpty()) {
        return false;
    }
    quint8 expected = (quint8)data.at(size);
    quint8 got = sioChecksum(data, size);
    if (expected == got) {
        data.resize(size);
        return data;
    } else {
        if (verbose) {
            emitUiMessage(tr("Data frame checksum error, expected: %1, got: %2. (%3)")
                           .arg(expected)
                           .arg(got)
                           .arg(QString(data.toHex())), MessageType::WarningMessage);
        }
        data.clear();
        return data;
    }
}

bool StandardSerialPortBackend::writeDataFrame(const QByteArray &data)
{
    QByteArray copy(data);
    copy.resize(copy.size() + 1);
    copy[copy.size() - 1] = sioChecksum(copy, copy.size() - 1);
    SioWorker::usleep(50);
    return writeRawFrame(copy);
}

bool StandardSerialPortBackend::writeCommandAck()
{
    return writeRawFrame(QByteArray(1, 65));
}

bool StandardSerialPortBackend::writeCommandNak()
{
    return writeRawFrame(QByteArray(1, 78));
}

bool StandardSerialPortBackend::writeDataAck()
{
    return writeRawFrame(QByteArray(1, 65));
}

bool StandardSerialPortBackend::writeDataNak()
{
    return writeRawFrame(QByteArray(1, 78));
}

bool StandardSerialPortBackend::writeComplete()
{
    SioWorker::usleep(300);
    return writeRawFrame(QByteArray(1, 67));
}

bool StandardSerialPortBackend::writeError()
{
    SioWorker::usleep(300);
    return writeRawFrame(QByteArray(1, 69));
}

quint8 StandardSerialPortBackend::sioChecksum(const QByteArray &data, uint size)
{
    uint i;
    uint sum = 0;

    for (i=0; i < size; i++) {
        sum += (quint8)data.at(i);
        if (sum > 255) {
            sum -= 255;
        }
    }

    return sum;
}

QByteArray StandardSerialPortBackend::readRawFrame(uint size, bool verbose)
{
    QByteArray data;
    int result;
    uint total, rest;

    data.resize(size);

    total = 0;
    rest = size;
    QTime startTime = QTime::currentTime();
    int timeOut = data.count() * 12000 / mSpeed + 10;
    int elapsed;

    do {
        result = ::read(mHandle, data.data() + total, rest);
        if (result < 0 && errno != EAGAIN) {
            emitUiMessage(tr("Cannot read from serial port: %1")
                           .arg(lastErrorMessage()), MessageType::ErrorMessage);
            data.clear();
            return data;
        }
        if (result < 0) {
            result = 0;
        }
        total += result;
        rest -= result;
        elapsed = QTime::currentTime().msecsTo(startTime);
    } while (total < size && elapsed > -timeOut);

    if ((uint)total != size) {
        if (verbose) {
            data.resize(total);
            emitUiMessage(tr("Serial port read timeout."), MessageType::ErrorMessage);
        }
        data.clear();
        return data;
    }
    return data;
}

bool StandardSerialPortBackend::writeRawFrame(const QByteArray &data)
{
    int result;
    uint total, rest;

    total = 0;
    rest = data.count();
    QTime startTime = QTime::currentTime();
    int timeOut = data.count() * 12000 / mSpeed + 10;
    int elapsed;

    if (tcdrain(mHandle) != 0) {
        emitUiMessage(tr("Cannot flush serial port write buffer: %1")
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }

    do {
        result = ::write(mHandle, data.constData() + total, rest);
        if (result < 0 && errno != EAGAIN) {
            emitUiMessage(tr("Cannot read from serial port: %1")
                           .arg(lastErrorMessage()), MessageType::ErrorMessage);
            return false;
        }
        if (result < 0) {
            result = 0;
        }
        total += result;
        rest -= result;
        elapsed = QTime::currentTime().msecsTo(startTime);
    } while (total < (uint)data.count() && elapsed > -timeOut);

    if (total != (uint)data.count()) {
        emitUiMessage(tr("Serial port write timeout."), MessageType::ErrorMessage);
        return false;
    }

    return true;
}

QString StandardSerialPortBackend::lastErrorMessage()
{
    return QString(strerror(errno)) + ".";
}

AtariSioBackend::AtariSioBackend(QObject *parent)
    : AbstractSerialPortBackend(parent)
{
    mHandle = -1;
}

AtariSioBackend::~AtariSioBackend()
{
    if (isOpen()) {
        close();
    }
}

QString AtariSioBackend::defaultPortName()
{
    return QString("/dev/atarisio0");
}

bool AtariSioBackend::open(QString name, SerialLine method)
{
    if (isOpen()) {
        close();
    }

    mHandle = ::open(name.toLocal8Bit().constData(), O_RDWR);

    if (mHandle < 0) {
        emitUiMessage(tr("Cannot open serial port '%1': %2")
                       .arg(name, lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }

    int version;
    version = ioctl(mHandle, ATARISIO_IOC_GET_VERSION);

    if (version < 0) {
        emitUiMessage(tr("Cannot open AtariSio driver '%1': Cannot determine AtariSio version.")
                      .arg(name), MessageType::ErrorMessage);
        close();
        return false;
    }

    if ((version >> 8) != (ATARISIO_VERSION >> 8) ||
         (version & 0xff) < (ATARISIO_VERSION & 0xff)) {
        emitUiMessage(tr("Cannot open AtariSio driver '%1': Incompatible AtariSio version.")
                      .arg(name), MessageType::ErrorMessage);
        close();
        return false;
    }

    int mode;

    switch (method) {
    case Ri:
        mode = ATARISIO_MODE_SIOSERVER;
        break;
    case Dsr:
        mode = ATARISIO_MODE_SIOSERVER_DSR;
        break;
    default:
    case Cts:
        mode = ATARISIO_MODE_SIOSERVER_CTS;
        break;
    }

    if (ioctl(mHandle, ATARISIO_IOC_SET_MODE, mode) < 0) {
        emitUiMessage(tr("Cannot set AtariSio driver mode: %1")
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
        close();
        return false;
    }

    if (ioctl(mHandle, ATARISIO_IOC_SET_AUTOBAUD, true) < 0) {
        emitUiMessage(tr("Cannot set AtariSio to autobaud mode: %1")
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
        close();
        return false;
    }

    mMethod = method;

    pipe(mCancelHandles);

    return true;
}

bool AtariSioBackend::isOpen()
{
    return mHandle >= 0;
}

void AtariSioBackend::close()
{
    cancel();
    if (::close(mHandle)) {
        emitUiMessage(tr("Cannot close serial port: %1")
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
    }
    ::close(mCancelHandles[0]);
    ::close(mCancelHandles[1]);
    mHandle = -1;
}

void AtariSioBackend::cancel()
{
    ::write(mCancelHandles[1], "C", 1);
}

void AtariSioBackend::setSpeed(uint speed)
{
    if (ioctl(mHandle, ATARISIO_IOC_SET_BAUDRATE, speed) < 0) {
        emitUiMessage(QString("Cannot set AtariSio speed to %1: %2").arg(speed).arg(lastErrorMessage()), MessageType::ErrorMessage);
    } else {
        emit speedChanged(speed);
        emitUiMessage(QString("Serial port speed set to %1.").arg(speed), MessageType::ImportantMessage);
        mSpeed = speed;
    }
}

uint AtariSioBackend::speed()
{
    return mSpeed;
}

QByteArray AtariSioBackend::readCommandFrame()
{
    QByteArray data;

    fd_set read_set;
    fd_set except_set;

    int ret;

    int maxfd = mHandle;

    FD_ZERO(&except_set);
    FD_SET(mHandle, &except_set);

    if (mCancelHandles[0] > maxfd) {
        maxfd = mCancelHandles[0];
    }
    FD_ZERO(&read_set);
    FD_SET(mCancelHandles[0], &read_set);

    ret = select(maxfd+1, &read_set, NULL, &except_set, 0);
    if (ret == -1 || ret == 0) {
        return data;
    }
    if (FD_ISSET(mCancelHandles[0], &read_set)) {
        return data;
    }
    if (FD_ISSET(mHandle, &except_set)) {
        SIO_command_frame frame;
        if (ioctl(mHandle, ATARISIO_IOC_GET_COMMAND_FRAME, &frame) < 0) {
            return data;
        }
        data.resize(4);
        data[0] = frame.device_id;
        data[1] = frame.command;
        data[2] = frame.aux1;
        data[3] = frame.aux2;

        int sp = ioctl(mHandle, ATARISIO_IOC_GET_BAUDRATE);
        if (sp >= 0 && mSpeed != sp) {
            mSpeed = sp;
            emit speedChanged(mSpeed);
            emitUiMessage(QString("Serial port speed set to %1.").arg(mSpeed), MessageType::ImportantMessage);
        }

        return data;
    }
    emitUiMessage("Illegal condition using select!", MessageType::ErrorMessage);
    return data;
}

QByteArray AtariSioBackend::readDataFrame(uint size, bool verbose)
{
    QByteArray data;
    SIO_data_frame frame;

    data.resize(size);
    frame.data_buffer = (unsigned char*)data.data();
    frame.data_length = size;

    if (ioctl(mHandle, ATARISIO_IOC_RECEIVE_DATA_FRAME, &frame) < 0 ) {
        if (verbose) {
            emitUiMessage(tr("Cannot read data frame: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        }
        data.clear();
    }

    return data;
}

bool AtariSioBackend::writeDataFrame(const QByteArray &data)
{
    SIO_data_frame frame;

    frame.data_buffer = (unsigned char*)data.constData();
    frame.data_length = data.size();

    if (ioctl(mHandle, ATARISIO_IOC_SEND_DATA_FRAME, &frame) < 0 ) {
        emitUiMessage(tr("Cannot write data frame: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }

    return true;
}

bool AtariSioBackend::writeCommandAck()
{
    if (ioctl(mHandle, ATARISIO_IOC_SEND_COMMAND_ACK) < 0 ) {
        if (errno != EATARISIO_COMMAND_TIMEOUT) {
            emitUiMessage(tr("Cannot write command ACK: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        }
        return false;
    }
    return true;
}

bool AtariSioBackend::writeCommandNak()
{
    if (ioctl(mHandle, ATARISIO_IOC_SEND_COMMAND_NAK) < 0 ) {
        emitUiMessage(tr("Cannot write command NAK: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    return true;
}

bool AtariSioBackend::writeDataAck()
{
    if (ioctl(mHandle, ATARISIO_IOC_SEND_DATA_ACK) < 0 ) {
        emitUiMessage(tr("Cannot write data ACK: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    return true;
}

bool AtariSioBackend::writeDataNak()
{
    if (ioctl(mHandle, ATARISIO_IOC_SEND_DATA_NAK) < 0 ) {
        emitUiMessage(tr("Cannot write data NAK: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    return true;
}

bool AtariSioBackend::writeComplete()
{
    if (ioctl(mHandle, ATARISIO_IOC_SEND_COMPLETE) < 0 ) {
        emitUiMessage(tr("Cannot write COMPLETE byte: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    return true;
}

bool AtariSioBackend::writeError()
{
    if (ioctl(mHandle, ATARISIO_IOC_SEND_ERROR) < 0 ) {
        emitUiMessage(tr("Cannot write ERROR byte: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    return true;
}

QString AtariSioBackend::lastErrorMessage()
{
    switch (errno) {
        case EATARISIO_ERROR_BLOCK_TOO_LONG:
            return tr("Block too long.");
            break;
        case EATARISIO_COMMAND_NAK:
            return tr("Command not acknowledged.");
            break;
        case EATARISIO_COMMAND_TIMEOUT:
            return tr("Command timeout.");
            break;
        case EATARISIO_CHECKSUM_ERROR:
            return tr("Checksum error.");
            break;
        case EATARISIO_COMMAND_COMPLETE_ERROR:
            return tr("Device error.");
            break;
        case EATARISIO_DATA_NAK:
            return tr("Data frame not acknowledged.");
            break;
        case EATARISIO_UNKNOWN_ERROR:
            return tr("Unknown AtariSio driver error.");
            break;
        default:
            return QString(strerror(errno)) + ".";
    }
}
