#include "serialport-win32.h"
#include "sioworker.h"

#include <string.h>

AbstractSerialPortBackend::AbstractSerialPortBackend(QObject *parent)
    : QObject(parent)
{
}

AbstractSerialPortBackend::~AbstractSerialPortBackend()
{
}

StandardSerialPortBackend::StandardSerialPortBackend(QObject *parent)
    : AbstractSerialPortBackend(parent)
{
    mHandle = INVALID_HANDLE_VALUE;
}

StandardSerialPortBackend::~StandardSerialPortBackend()
{
    if (isOpen()) {
        close();
    }
}

QString StandardSerialPortBackend::StandardSerialPortBackend::defaultPortName()
{
    return QString("COM1");
}

bool StandardSerialPortBackend::open(QString name, SerialLine method)
{
    if (isOpen()) {
        close();
    }
    mHandle = (CreateFile(
        (WCHAR*)name.utf16(),
        GENERIC_READ | GENERIC_WRITE,
        0,
        NULL,
        OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
        0
    ));
    if (mHandle == INVALID_HANDLE_VALUE) {
        emitUiMessage(tr("Cannot open serial port '%1': %2").arg(name, lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    if (!EscapeCommFunction(mHandle, CLRRTS)) {
        emitUiMessage(tr("Cannot clear RTS line in serial port '%1': %2").arg(name, lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }
    if (!EscapeCommFunction(mHandle, CLRDTR)) {
        emitUiMessage(tr("Cannot clear DTR line in serial port '%1': %2").arg(name, lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }

    mMethod = method;

    mCanceled = false;

    mCancelHandle = CreateEvent(0, true, false, 0);

    return true;
}

bool StandardSerialPortBackend::isOpen()
{
    return mHandle != INVALID_HANDLE_VALUE;
}

void StandardSerialPortBackend::close()
{
    cancel();
    if (!CloseHandle(mHandle)) {
        emitUiMessage(tr("Cannot close serial port: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
    }
    CloseHandle(mCancelHandle);
    mCancelHandle = mHandle = INVALID_HANDLE_VALUE;
}

void StandardSerialPortBackend::cancel()
{
    mCanceled = true;
    SetEvent(mCancelHandle);
}

void StandardSerialPortBackend::setSpeed(uint speed)
{
    DCB dcb;
    COMMTIMEOUTS to;

    /* Adjust parameters */
    dcb.DCBlength = sizeof dcb;
    dcb.BaudRate = speed;
    dcb.fBinary = TRUE;
    dcb.fParity = FALSE;

    dcb.fOutxCtsFlow = FALSE;
    dcb.fOutxDsrFlow = FALSE;
    dcb.fDtrControl = DTR_CONTROL_DISABLE;
    dcb.fDsrSensitivity = FALSE;
    dcb.fTXContinueOnXoff = FALSE;
    dcb.fOutX = FALSE;
    dcb.fInX = FALSE;
    dcb.fErrorChar = FALSE;
    dcb.fNull = FALSE;
    dcb.fRtsControl = RTS_CONTROL_DISABLE;
    dcb.fAbortOnError = FALSE;
    dcb.fDummy2 = 0;
    dcb.wReserved = 0;
    dcb.XonLim = 0;
    dcb.XoffLim = 0;
    dcb.ByteSize = 8;
    dcb.Parity = NOPARITY;
    dcb.StopBits = ONESTOPBIT;
    dcb.XonChar = 0;
    dcb.XoffChar = 0;
    dcb.ErrorChar = 0;
    dcb.EofChar = 0;
    dcb.EvtChar = 0;
    dcb.wReserved1 = 0;

    /* Set serial port state */
    if (!SetCommState(mHandle, &dcb)) {
        emitUiMessage(tr("Cannot set serial port speed to %1: %2")
                       .arg(speed)
                       .arg(lastErrorMessage()), MessageType::ErrorMessage);
        return;
    }

    /* Adjust serial port timeouts */
    to.ReadIntervalTimeout = 0;
    to.ReadTotalTimeoutMultiplier = 10;
    to.ReadTotalTimeoutConstant = 0;
    to.WriteTotalTimeoutMultiplier = 10;
    to.WriteTotalTimeoutConstant = 0;

    /* Set serial port timeouts */
    if (!SetCommTimeouts(mHandle, &to)) {
        emitUiMessage(tr("Cannot set serial port timeouts: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return;
    }

    emit speedChanged(speed);
    emitUiMessage(QString("Serial port speed set to %1.").arg(speed), MessageType::ImportantMessage);
    mSpeed = speed;
    return;
}

uint StandardSerialPortBackend::speed()
{
    return mSpeed;
}

QByteArray StandardSerialPortBackend::readCommandFrame()
{
    QByteArray data;
    DWORD mask;

    switch (mMethod) {
    case Ri:
        mask = EV_RING;
        break;
    case Dsr:
        mask = EV_DSR;
        break;
    default:
    case Cts:
        mask = EV_CTS;
        break;
    }

    if (!SetCommMask(mHandle, mask)) {
        emitUiMessage(tr("Cannot set serial port event mask: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return data;
    }

    data.clear();
    OVERLAPPED ov;

    memset(&ov, 0, sizeof(ov));
    ov.hEvent = CreateEvent(0, true, false, 0);

    HANDLE events[2];
    events[0] = ov.hEvent;
    events[1] = mCancelHandle;

    if (!WaitCommEvent(mHandle, &mask, &ov)) {
        if (GetLastError() == ERROR_IO_PENDING) {
            DWORD x = WaitForMultipleObjects(2, events, false, INFINITE);
            CloseHandle(ov.hEvent);
            if (x == WAIT_OBJECT_0 + 1) {
                data.clear();
                return data;
            }
            if (x == WAIT_FAILED) {
                emitUiMessage(tr("Cannot wait for serial port event: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
                data.clear();
                return data;
            }
        } else {
            CloseHandle(ov.hEvent);
            emitUiMessage(tr("Cannot wait for serial port event: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
            return data;
        }
    }

    if (!PurgeComm(mHandle, PURGE_RXCLEAR)) {
        emitUiMessage(tr("Cannot clear serial port read buffer: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return data;
    }

    data = readDataFrame(4, false);

    if (!data.isEmpty()) {
        do {
            GetCommModemStatus(mHandle, &mask);
        } while (mask && !mCanceled);
    }

    return data;
}

QByteArray StandardSerialPortBackend::readDataFrame(uint size, bool verbose)
{
    QByteArray data = readRawFrame(size + 1, verbose);
    if (data.isEmpty()) {
        return false;
    }
    quint8 expected = (quint8)data.at(size);
    quint8 got = sioChecksum(data, size);
    if (expected == got) {
        data.resize(size);
        return data;
    } else {
        if (verbose) {
            emitUiMessage(tr("Data frame checksum error, expected: %1, got: %2. (%3)")
                           .arg(expected)
                           .arg(got)
                           .arg(QString(data.toHex())), MessageType::WarningMessage);
        }
        data.clear();
        return data;
    }
}

bool StandardSerialPortBackend::writeDataFrame(const QByteArray &data)
{
    QByteArray copy(data);
    copy.resize(copy.size() + 1);
    copy[copy.size() - 1] = sioChecksum(copy, copy.size() - 1);
    SioWorker::usleep(50);
    return writeRawFrame(copy);
}

bool StandardSerialPortBackend::writeCommandAck()
{
    return writeRawFrame(QByteArray(1, 65));
}

bool StandardSerialPortBackend::writeCommandNak()
{
    return writeRawFrame(QByteArray(1, 78));
}

bool StandardSerialPortBackend::writeDataAck()
{
    return writeRawFrame(QByteArray(1, 65));
}

bool StandardSerialPortBackend::writeDataNak()
{
    return writeRawFrame(QByteArray(1, 78));
}

bool StandardSerialPortBackend::writeComplete()
{
    SioWorker::usleep(300);
    return writeRawFrame(QByteArray(1, 67));
}

bool StandardSerialPortBackend::writeError()
{
    SioWorker::usleep(300);
    return writeRawFrame(QByteArray(1, 69));
}

quint8 StandardSerialPortBackend::sioChecksum(const QByteArray &data, uint size)
{
    uint i;
    uint sum = 0;

    for (i=0; i < size; i++) {
        sum += (quint8)data.at(i);
        if (sum > 255) {
            sum -= 255;
        }
    }

    return sum;
}

QByteArray StandardSerialPortBackend::readRawFrame(uint size, bool verbose)
{
    QByteArray data;
    DWORD result;
    OVERLAPPED ov;

    memset(&ov, 0, sizeof(ov));
    ov.hEvent = CreateEvent(0, true, false, 0);

    if (ov.hEvent == INVALID_HANDLE_VALUE) {
        emitUiMessage(tr("Cannot create event: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return data;
    }

    data.resize(size);
    if (!ReadFile(mHandle, data.data(), size, &result, &ov)) {
        if (GetLastError() == ERROR_IO_PENDING) {
            if (!GetOverlappedResult(mHandle, &ov, &result, true)) {
                emitUiMessage(tr("Cannot read from serial port: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
                data.clear();
                CloseHandle(ov.hEvent);
                return data;
            }
        } else {
            emitUiMessage(tr("Cannot read from serial port: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
            data.clear();
            CloseHandle(ov.hEvent);
            return data;
        }
    }
    CloseHandle(ov.hEvent);
    if (verbose && result != (DWORD)size) {
        emitUiMessage(tr("Serial port read timeout."), MessageType::ErrorMessage);
        data.clear();
        return data;
    }
    return data;
}

bool StandardSerialPortBackend::writeRawFrame(const QByteArray &data)
{
    DWORD result;
    OVERLAPPED ov;

    memset(&ov, 0, sizeof(ov));
    ov.hEvent = CreateEvent(0, true, false, 0);

    if (!PurgeComm(mHandle, PURGE_TXCLEAR)) {
        emitUiMessage(tr("Cannot clear serial port write buffer: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
        return false;
    }

    if (!WriteFile(mHandle, data.constData(), data.size(), &result, &ov)) {
        if (GetLastError() == ERROR_IO_PENDING) {
            if (!GetOverlappedResult(mHandle, &ov, &result, true)) {
                emitUiMessage(tr("Cannot write to serial port: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
                CloseHandle(ov.hEvent);
                return false;
            }
        } else {
            emitUiMessage(tr("Cannot write to serial port: %1").arg(lastErrorMessage()), MessageType::ErrorMessage);
            return false;
        }
    }
    CloseHandle(ov.hEvent);
    if (result != (DWORD)data.size()) {
        emitUiMessage(tr("Serial port write timeout."), MessageType::ErrorMessage);
        return false;
    }
    return true;
}

QString StandardSerialPortBackend::lastErrorMessage()
{
    QString result;
    LPVOID lpMsgBuf;

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    result.setUtf16((ushort *)lpMsgBuf, wcslen((wchar_t*)lpMsgBuf) - 2);
    LocalFree(lpMsgBuf);
    return result;
}
