#include "sioworker.h"

#include "aspeqtsettings.h"

SioDevice::SioDevice(SioWorker *worker)
        :QObject()
{
    sio = worker;
    m_deviceNo = -1;
    connect(this, SIGNAL(uiMessage(const QString, MessageType::UiMessageType)), sio, SIGNAL(uiMessage(QString, MessageType::UiMessageType)));
}

SioDevice::~SioDevice()
{
    uiMessage(tr("Freeing %1")
              .arg(m_deviceNo), MessageType::DebugMessage);
    if (m_deviceNo != -1) {
        sio->uninstallDevice(m_deviceNo);
    }
}

QString SioDevice::deviceName()
{
    return sio->deviceName(m_deviceNo);
}

SioWorker::SioWorker()
        : QThread()
{
    deviceMutex = new QMutex(QMutex::Recursive);
    for (int i=0; i <= 255; i++) {
        devices[i] = 0;
    }
    mPort = 0;
}

SioWorker::~SioWorker()
{
    for (int i=0; i <= 255; i++) {
        if (devices[i]) {
            delete devices[i];
        }
    }
    delete deviceMutex;
}

bool SioWorker::wait(unsigned long time)
{
    mustTerminate = true;
    if (mPort) {
        mPort->cancel();
    }
    return QThread::wait(time);
    if (mPort) {
        delete mPort;
        mPort = 0;
    }
}

void SioWorker::start(Priority p)
{
#ifdef Q_OS_LINUX
    if (aspeqtSettings->backend() == 0) {
        mPort = new StandardSerialPortBackend(0);
    } else {
        mPort = new AtariSioBackend(0);
    }
#else
    mPort = new StandardSerialPortBackend(0);
#endif
    mustTerminate = false;
    QThread::start(p);
}

void SioWorker::run()
{
    connect(mPort, SIGNAL(uiMessage(const QString, MessageType::UiMessageType)), this, SIGNAL(uiMessage(QString, MessageType::UiMessageType)));
    connect(mPort, SIGNAL(speedChanged(uint)), this, SIGNAL(speedChanged(uint)));
    qRegisterMetaType<MessageType::UiMessageType>("MessageType::UiMessageType");

    int method = aspeqtSettings->handshakingMethod();
    maxSpeed = aspeqtSettings->maximumSerialPortSpeed();
    AbstractSerialPortBackend::SerialLine handshake;
    QString handshakeString;

    switch(method) {
        case 1:
            handshake = AbstractSerialPortBackend::Dsr;
            handshakeString = "DSR";
            break;
        case 2:
            handshake = AbstractSerialPortBackend::Cts;
            handshakeString = "CTS";
            break;
        default:
            handshake = AbstractSerialPortBackend::Ri;
            handshakeString = "RI";
            break;
    }

    /* Open serial port */
    QString portName;
#ifdef Q_OS_LINUX
    if (aspeqtSettings->backend() == 0) {
        portName = aspeqtSettings->serialPortName();
    } else {
        portName = aspeqtSettings->atariSioDriverName();
    }
#else
    portName = aspeqtSettings->serialPortName();
#endif
    if (!mPort->open(portName, handshake)) {
        return;
    }

    /* Set serial mPort speed */
    mPort->setSpeed(19200);

    if (mPort->speed() != 19200) {
        mPort->close();
        return;
    }

    /* Notify the user that emulation is started */
    emitUiMessage(tr("Emulation started on '%1' with %2 handshaking.").arg(portName, handshakeString), MessageType::ImportantMessage);

    /* Process SIO commands until we're explicitly stopped */
    int retries = 0;
    while (!mustTerminate) {
        QByteArray cmd = mPort->readCommandFrame();
        if (mustTerminate) {
            break;
        }
        if (cmd.isEmpty()) {
            if (retries < 3) {
                retries++;
                continue;
            }
            retries = 0;
            uint newSpeed;
            uint oldSpeed = mPort->speed();
            if (oldSpeed == 19200) {
                switch (maxSpeed) {
                    case 0:
                        newSpeed = 19200;
                        break;
                    case 1:
                        newSpeed = 38400;
                        break;
                    case 2:
                    default:
                        newSpeed = 57600;
                        break;
                }
            } else {
                newSpeed = 19200;
            }
            mPort->setSpeed(newSpeed);
            if (mPort->speed() != newSpeed) {
                mPort->close();
                return;
            }
            continue;
        }
        /* Decode the command */
        quint8 no = (quint8)cmd[0];
        quint8 command = (quint8)cmd[1];
        quint16 aux = ((quint8)cmd[2]) + ((quint8)cmd[3] * 256);

        /* Redirect the command to the appropriate device */
        deviceMutex->lock();
        if (devices[no]) {
            devices[no]->lock();
            devices[no]->handleCommand(command, aux);
            devices[no]->unlock();
        } else {
            emitUiMessage(QString("[%1] command: $%2, aux: $%3 ignored.")
                           .arg(deviceName(no))
                           .arg(command, 2, 16, QChar('0'))
                           .arg(aux, 4, 16, QChar('0')), MessageType::UnimportantMessage);
        }
        deviceMutex->unlock();
        cmd.clear();
    }
    mPort->close();
}

void SioWorker::installDevice(quint8 no, SioDevice *device)
{
    deviceMutex->lock();
    if (devices[no]) {
        delete devices[no];
    }
    devices[no] = device;
    device->setDeviceNo(no);
    deviceMutex->unlock();
}

void SioWorker::uninstallDevice(quint8 no)
{
    deviceMutex->lock();
    if (devices[no]) {
        devices[no]->setDeviceNo(-1);
    }
    devices[no] = 0;
    deviceMutex->unlock();
}

void SioWorker::swapDevices(quint8 d1, quint8 d2)
{
    SioDevice *t1, *t2;

    deviceMutex->lock();
    t1 = devices[d1];
    t2 = devices[d2];
    uninstallDevice(d1);
    uninstallDevice(d2);
    if (t2) {
        installDevice(d1, t2);
    }
    if (t1) {
        installDevice(d2, t1);
    }
    deviceMutex->unlock();
}

SioDevice* SioWorker::getDevice(quint8 no)
{
    SioDevice *result;
    deviceMutex->lock();
    result = devices[no];
    deviceMutex->unlock();
    return result;
}

QString SioWorker::deviceName(quint8 device)
{
    QString result;
    switch (device) {
        case 0x31:
        case 0x32:
        case 0x33:
        case 0x34:
        case 0x35:
        case 0x36:
        case 0x37:
        case 0x38:
            result = tr("Disk %1").arg(((uint)device) - 0x30);
            break;
        case 0x45:
            result = tr("APE time downloader");
            break;
        default:
            result = tr("Device $%1").arg(((uint)device), 2, 16, QChar('0'));
            break;
    }
    return result;
}
